/**
 * @file  bidiPIXUS850i.c
 * @brief Impletation of Bi-di Lib module (canon pixus 850i)
 *
 * @date 2004/02/12
 * @version 1.0.0
 *
 * Copyright (c) 2002-2004, BBR Co.Ltd., All rights reserved.
 */
/*
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#define _GNU_SOURCE     /* For type "sighandler_t" */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include "bidiPlugin.h"
#include "bidiDbg.h"
#include "bscc2sts.h"
#include "sts2XML.h"
#include "bscc2XML.h"
#include "bidiLib.h"


/** åƥ */
static jmp_buf env;

/**
 * longjmp ԤäæФ¹Ԥ뤿 signal handler
 *
 * @param None.
 * @return None.
 */
static void sigExit(void)
{
    longjmp(env, 1);
}


/**
 * Create a new Bi-di Object.
 *
 * @param fdRead  File descriptor for reading the printer status data from 
 *                the Bi-di Plug-in module.
 * @param fdWrite File descriptor for writing the printer command data to 
 *                the Bi-di Plug-in module.
 * @param pURI    Device URI.
 * @return Pointer to the Bi-di Object.
 */
void *bidiLibNew(int fdRead, int fdWrite, char *pURI)
{
    BidiLibC *p;
    char *uri;

    // BidiLibC ΰƤ
    p = (BidiLibC *) malloc(sizeof(BidiLibC));
    if (p == NULL) {
        _dbgPrint(sys_errlist[errno]);
        return NULL;
    }

    // pURI Ƥ򥳥ԡ
    if (pURI != NULL) {
        char *uri = strdup(pURI);
        if (uri == NULL) {
            _dbgPrint(sys_errlist[errno]);
            free(p);
            return NULL;
        }
    } else {
        uri = NULL;
    }

    // BidiLibC 
    p->fdRead = fdRead;
    p->fdWrite = fdWrite;
    p->uri = uri;
    p->isPrinterUp = 0;
    p->idJob = 0;
    p->xmlSize = 0;
    p->pXmlBuf = 0;

    return p;
}

/**
 * Destroy the Bi-di Object.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @return None.
 */
void bidiLibDestroy(void *pBidiLibC)
{
    // Ƥΰγ
    free(((BidiLibC *) pBidiLibC)->uri);
    free(pBidiLibC);
}

/**
 * Get the Bi-di Object capabilities.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @param cap       Enum of the Bi-di Object capabilities.
 * @retval BIDI_TRUE  The condition specified by cap is true.
 * @retval BIDI_FALSE The condition specified by cap is false.
 * @retval BIDI_ERROR When error.
 */
int bidiLibGetCap(void *pBidiLibC, BidiCap cap)
{
    switch (cap) {
    case BIDI_CAP_WRITE:
        return BIDI_FALSE;
    case BIDI_CAP_JOB:
        return BIDI_TRUE;
    case BIDI_CAP_CTRL:
        return BIDI_FALSE;
    default:
        _dbgPrint("illegal cap case.");
        return BIDI_ERROR;
    }
}

/**
 * Start the printing job sequence.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @param idJob     Job ID.
 * @retval BIDI_OK  Normal terminated.
 */
int bidiLibStartJob(void *pBidiLibC, int idJob)
{
    ((BidiLibC *) pBidiLibC)->isPrinterUp = 0;
    ((BidiLibC *) pBidiLibC)->idJob = idJob;
    return BIDI_OK;
}

/**
 * Stop the printing job sequence.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @retval BIDI_OK        Normal terminated.
 * @retval BIDI_ERROR     When error.
 * @retval BIDI_EPROGRESS When the job can't stop(the job is running).
 */
int bidiLibEndJob(void *pBidiLibC)
{
    int ret;
    int stat;
    bscc2sts_tbl sts;

    int fdRead = ((BidiLibC *) pBidiLibC)->fdRead;
    int fdWrite = ((BidiLibC *) pBidiLibC)->fdWrite;

    ret = genSts(PRINTER_NAME, fdRead, fdWrite, &sts);
    if (ret != 0) {
        return BIDI_ERROR;
    }

    // ֤ˤʤä isPrinterUp  true ˤ
    stat = isPrinting(&sts);
    if (stat && !((BidiLibC *) pBidiLibC)->isPrinterUp) {
        ((BidiLibC *) pBidiLibC)->isPrinterUp = 1;
    }
    if (stat || !((BidiLibC *) pBidiLibC)->isPrinterUp) {
        return BIDI_EPROGRESS;
    } else {
        return BIDI_OK;
    }
}

/**
 * Cancel the printing job sequence.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @param idJob     Job ID.
 * @retval BIDI_EPROGRESS When the job can't stop(the job is running).
 * @retval BIDI_ENDJOB    When the job is not exist in Bi-di Plug-in module.
 */
int bidiLibCancelJob(void *pBidiLibC, int idJob)
{
    if (((BidiLibC *) pBidiLibC)->idJob != idJob) {
        return BIDI_ENOJOB;
    }
    return BIDI_EPROGRESS;
}

/**
 * Get the file descriptor for reading from the Bi-di Object.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @return File descriptor.
 */
int bidiLibGetReadFD(void *pBidiLibC)
{
    return ((BidiLibC *) pBidiLibC)->fdRead;
}

/**
 * Start the reading sequence.
 *
 * @param pBidiLibC  Pointer to the Bi-di Object.
 * @param idReadMode Enum of the Bi-di Object reading mode.
 * @param pLang      Pointer to the language string.
 *                   (ʸƱͤθ+ϰ+ʸɻʸ)
 *                   NULL able.
 * @retval BIDI_OK        Normal terminated.
 * @retval BIDI_ERROR     When error.
 */
int bidiLibStartRead(void *pBidiLibC, BidiReadMode idReadMode, char *pLang)
{
    int ret;
    int stat;
    bscc2sts_tbl sts;
    BidiLibC *p = (BidiLibC *) pBidiLibC;

    // Read ModeΥå
    if (idReadMode != BIDI_READ_PRT_MIB_SUMMARY) {
        return BIDI_ERROR;
    }

    // stsǡ
    ret = genSts(PRINTER_NAME, p->fdRead, p->fdWrite, &sts);
    if (ret != 0) {
        return BIDI_ERROR;
    }

    // XMLǡι¤ΤؤΥå
    ret = sts2XML(&sts, p->xmlBuf, XML_BUFFER_SIZE, pLang);
    if (ret == -1) {
        return BIDI_ERROR;
    }
    p->xmlSize = strlen(p->xmlBuf);
    p->pXmlBuf = 0;

    // ֤ˤʤä isPrinterUp  true ˤ
    stat = isPrinting(&sts);
    if (stat && !((BidiLibC *) pBidiLibC)->isPrinterUp) {
        ((BidiLibC *) pBidiLibC)->isPrinterUp = 1;
    }

    return BIDI_OK;
}

/**
 * Read the printer status from the Bi-di Object.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @param pBuf      Pointer to the buffer for storing the printer status data.
 * @param nBufBytes Number of bytes of the printer status data buffer.
 * @retval >0           Number of bytes of the printer status data read.
 * @retval 0            When no data remains.
 * @retval BIDI_ERROR   When error.
 * @retval BIDI_EINTR   When recieve signal.(read 0byte)
 */
int bidiLibRead(void *pBidiLibC, void *pBuf, int nBufBytes)
{
    BidiLibC *p = (BidiLibC *) pBidiLibC;
    int i = 0;
    int len;
    sighandler_t oldHandler;   // ʥϥɥΤѿ

    // åƥȤ򥻡
    if (setjmp(env) == 0) {
        // longjmp Τ signal handler Υå
        oldHandler = signal(SIGALRM, (void *)sigExit);
        if (oldHandler == SIG_ERR) {
            _dbgPrint(sys_errlist[errno]);
            return BIDI_ERROR;
        }
    } else {
        // signal ȯΥƥ
        if (i == 0) {
            len = BIDI_EINTR;
        } else {
            p->pXmlBuf += i;
            len = i;
        }
        goto errRet;
    }

    int left, size;
    left = p->xmlSize - p->pXmlBuf;
    size = left > nBufBytes ? nBufBytes : left;
    for (i = 0; i < size; i++) {
        *((char *)pBuf + i) = p->xmlBuf[p->pXmlBuf + i];
    }
    p->pXmlBuf += size;
    len = size;
errRet:
    signal(SIGALRM, oldHandler);
    return len;
}

/**
 * Stop the reading sequence.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @retval BIDI_OK  Normal terminated.
 */
int bidiLibEndRead(void *pBidiLibC)
{
    ((BidiLibC *) pBidiLibC)->xmlSize = 0;
    ((BidiLibC *) pBidiLibC)->pXmlBuf = 0;
    return BIDI_OK;
}

/**
 * Get the file descriptor for writing to the Bi-di Object.
 *
 * @param pBidiLibC Pointer to the Bi-di Object.
 * @return File descriptor.
 */
int bidiLibGetWriteFD(void *pBidiLibC)
{
    return ((BidiLibC *) pBidiLibC)->fdWrite;
}
