/*
 *  OpenPrinting Vector Printer Driver for Canon Pixus850i
 *  Copyright (c) 2004 E&D Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
 */

/*
 * API InkJetPrinter Functions : Job Control
 */

/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"

/* BJ Print local include files */
#include "bjp_context.h"
#include "bjp_conf.h"
#include "bjp_draw.h"

/*==================================================================
  API Entry
===================================================================*/
/* Job Control Operations */
int Ijet_StartJob( int printerContext, char *jobInfo );
int Ijet_EndJob( int printerContext );
int Ijet_StartDoc( int printerContext, char *docInfo );
int Ijet_EndDoc( int printerContext );
int Ijet_StartPage( int printerContext, char *pageInfo );
int Ijet_EndPage( int printerContext );

/* Graphics State Object Operations */
int Ijet_QueryColorSpace( int printerContext, pdColorSpace *pcspace, int *pnum );
int Ijet_SetColorSpace( int printerContext, pdColorSpace cspace );
int Ijet_GetColorSpace( int printerContext, pdColorSpace *pcspace );

/* Bitmap Image Operations */
int Ijet_DrawImage( int printerContext, int sourceWidth, int sourceHeight, int colorDepth, pdImageFormat imageFormat, pdRectangle destinationSize, int count, void *imageData );
int Ijet_StartDrawImage( int printerContext, int sourceWidth, int sourceHeight, int colorDepth, pdImageFormat imageFormat, pdRectangle destinationSize );
int Ijet_TransferDrawImage( int printerContext, int count, void *imageData );
int Ijet_EndDrawImage( int printerContext );

/* Raster Image Operations */
int Ijet_StartRaster( int printerContext, int rasterWidth );
int Ijet_TransferRasterData( int printerContext, int count, unsigned char *data );
int Ijet_EndRaster( int printerContext );

int	(*Inkjet_Procs[])() =
{
	/* Creating and Managing Print Contexts */
	OpenPrinter,
	ClosePrinter,

	/* Job Control Operations */
	Ijet_StartJob,
	Ijet_EndJob,
	Ijet_StartDoc,
	Ijet_EndDoc,
	Ijet_StartPage,
	Ijet_EndPage,
	QueryDeviceCapability,
	QueryDeviceInfo,

	/* Graphics State Object Operations */
	NULL,	// Ijet_ResetCTM,
	NULL,	// Ijet_SetCTM,
	NULL,	// Ijet_GetCTM,
	NULL,	// Ijet_InitGS,
	NULL,	// Ijet_SaveGS,
	NULL,	// Ijet_RestoreGS,
	Ijet_QueryColorSpace,
	Ijet_SetColorSpace,
	Ijet_GetColorSpace,
	NULL,	// Ijet_QueryROP,
	NULL,	// Ijet_SetROP,
	NULL,	// Ijet_GetROP,
	NULL,	// Ijet_SetFillMode,
	NULL,	// Ijet_GetFillMode,
	NULL,	// Ijet_SetAlphaConstant,
	NULL,	// Ijet_GetAlphaConstant,
	NULL,	// Ijet_SetLineWidth,
	NULL,	// Ijet_GetLineWidth,
	NULL,	// Ijet_SetLineDash,
	NULL,	// Ijet_GetLineDash,
	NULL,	// Ijet_SetLineDashOffset,
	NULL,	// Ijet_GetLineDashOffset,
	NULL,	// Ijet_SetLineStyle,
	NULL,	// Ijet_GetLineStyle,
	NULL,	// Ijet_SetLineCap,
	NULL,	// Ijet_GetLineCap,
	NULL,	// Ijet_SetLineJoin,
	NULL,	// Ijet_GetLineJoin,
	NULL,	// Ijet_SetMiterLimit,
	NULL,	// Ijet_GetMiterLimit,
	NULL,	// Ijet_SetPaintMode,
	NULL,	// Ijet_GetPaintMode,
	NULL,	// Ijet_SetStrokeColor,
	NULL,	// Ijet_SetFillColor,
	NULL,	// Ijet_SetBgColor,

	/* Path Operations */
	NULL,	// Ijet_NewPath,
	NULL,	// Ijet_EndPath,
	NULL,	// Ijet_StrokePath,
	NULL,	// Ijet_FillPath,
	NULL,	// Ijet_StrokeFillPath,
	NULL,	// Ijet_SetClipPath,
	NULL,	// Ijet_ResetClipPath,
	NULL,	// Ijet_SetCurrentPoint,
	NULL,	// Ijet_LinePath,
	NULL,	// Ijet_PolygonPath,
	NULL,	// Ijet_RectanglePath,
	NULL,	// Ijet_RoundRectanglePath,
	NULL,	// Ijet_BezierPath,
	NULL,	// Ijet_ArchPath,

	/* Text Operations */
	NULL,	// Ijet_DrawBitmapText,

	/* Bitmap Image Operations */
	Ijet_DrawImage,
	Ijet_StartDrawImage,
	Ijet_TransferDrawImage,
	Ijet_EndDrawImage,

	/* Scan Line Operations */
	NULL,	// Ijet_StartScanline,
	NULL,	// Ijet_Scanline,
	NULL,	// Ijet_EndScanline,

	/* Raster Image Operations */
	Ijet_StartRaster,
	Ijet_TransferRasterData,
	NULL,	// Ijet_SkipRaster,
	Ijet_EndRaster,

	/* Stream Data Operations */
	NULL,	// Ijet_StartStream,
	NULL,	// Ijet_TransferStreamData,
	NULL,	// Ijet_EndStream,
};

/* support DeviceInfo */
static int	sdv_mediaCopy[]   = {
	1,
	-1,
	255,
	0 };
static int	sdv_orientation[] = {
	pdDev_orient_portrait,
	//pdDev_orient_landscape,
	0
};
static int	sdv_pageSizeID[]  = {
	pdDev_pgSize_a4,
	pdDev_pgSize_a5,
	pdDev_pgSize_b4,
	pdDev_pgSize_b5,
	pdDev_pgSize_letter,
	pdDev_pgSize_legal,
	0
};
static int	sdv_resolution[]  = {
	 600,  600,
	 300,  300,
	0
};
pd_supDevInfo	Inkjet_supDev = {
	sdv_mediaCopy,
	sdv_orientation,
	sdv_pageSizeID,	FALSE,
	sdv_resolution
};

/*==================================================================
  Local Functions
===================================================================*/
static void
setupControlInfo(pd_DeviceInfo	*devInfo,
		 BJPControlInfo *ctrlInfo)
{
	char *p = NULL;

	if(devInfo->mediaCopy > 0)
		ctrlInfo->NumCopies = devInfo->mediaCopy;
	switch(devInfo->pageSizeID){
	case pdDev_pgSize_a4:
		p = BJP_PAPERSIZE_A4; break;
	case pdDev_pgSize_a5:
		p = BJP_PAPERSIZE_A5; break;
	case pdDev_pgSize_b4:
		p = BJP_PAPERSIZE_B4; break;
	case pdDev_pgSize_b5:
		p = BJP_PAPERSIZE_B5; break;
	case pdDev_pgSize_letter:
		p = BJP_PAPERSIZE_LETTER; break;
	case pdDev_pgSize_legal:
		p = BJP_PAPERSIZE_LEGAL; break;
	}
	if(p)
		strcpy(ctrlInfo->MediaSizeName, p);
	if(devInfo->resolutionX > 0)
		ctrlInfo->resolutionX = devInfo->resolutionX;
	if(devInfo->resolutionY > 0)
		ctrlInfo->resolutionY = devInfo->resolutionY;
}

/* ----------------------------------------------------------------
 *
 *	prtDevInit
 *
 * ---------------------------------------------------------------- */
static bool
prtDevInit( pd_PrinterInfo *pri )
{
	BJPContext	*context;
	BJPControlInfo	ctrlInfo;

	memset(&ctrlInfo, 0, sizeof(BJPControlInfo));
	setupControlInfo(&pri->jobDev, &ctrlInfo);

	context = bjp_NewContext("PIXUS850I", pri->outputFD, &ctrlInfo);
	if(context == NULL){
		pd_DebugPrint("Error: newBJPContext()\n");
		errorno = pdFATALERROR;
		return(FALSE);
	}
	pd_DebugPrint("pdapi: context=%p\n", context);
	pri->printerData = (void*)context;
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	prtDevStart
 *
 * ---------------------------------------------------------------- */
static bool
prtDevStart(pd_PrinterInfo *pri,
	    pd_DeviceInfo *dev)
{
	BJPContext	*context = (BJPContext *)pri->printerData;

	bjp_StartJob(context);
	return( TRUE );
}
/* ----------------------------------------------------------------
 *
 *	prtDevEnd
 *
 * ---------------------------------------------------------------- */
static bool
prtDevEnd( pd_PrinterInfo *pri )
{
	BJPContext	*context = (BJPContext *)pri->printerData;

	bjp_EndJob(context);
	bjp_FreeContext(context);
	return( TRUE );
}

/*==================================================================
  API Functions
  ===================================================================*/

/* ----------------------------------------------------------------
 *
 *	Ijet_StartJob
 *
 * ---------------------------------------------------------------- */
int
Ijet_StartJob(int printerContext,
	      char *jobInfo)
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	if( jobInfo ) {
		pd_DebugPrint( "jobInfo = %s\n", jobInfo );
	}

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( pri->jobStarted ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Attribute Set */
	if( jobInfo && *jobInfo ) {
		if( !pd_SetJobInfo( pri, jobInfo ) )	return( pdERROR );
		memcpy( &pri->docDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );
		memcpy( &pri->pageDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );
	}
	/* Printer Device Initialize & Start */ 
	if( !prtDevInit( pri ) )	return( pdERROR );
	if( !prtDevStart( pri, &pri->jobDev ) )	return( pdERROR );
	memcpy( &pri->curDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );

	/* GraphicsState Initialize */
	pd_ClearGraphicsState( gs );
	pd_InitGraphicsState( gs );

	/* State Flag Set */
	pri->jobStarted = TRUE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_EndJob
 *
 * ---------------------------------------------------------------- */
int
Ijet_EndJob(int printerContext)
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted || pri->scnLinStarted || pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Printer Device End */ 
	if( !prtDevEnd( pri ) )	return( pdERROR );

	/* Attribute Set */
	memcpy( &pri->jobDev, &pri->defDev, sizeof( pd_DeviceInfo ) );
	memcpy( &pri->docDev, &pri->defDev, sizeof( pd_DeviceInfo ) );
	memcpy( &pri->pageDev, &pri->defDev, sizeof( pd_DeviceInfo ) );

	/* State Flag Set */
	pri->jobStarted = FALSE;
	pri->docStarted = FALSE;
	pri->pageStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_StartDoc
 *
 * ---------------------------------------------------------------- */
int
Ijet_StartDoc(int printerContext,
	      char *docInfo)
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	if( docInfo ) {
		pd_DebugPrint( "docInfo = %s\n", docInfo );
	}

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->docStarted ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Attribute Set */
	if( docInfo && *docInfo ) {
		if( !pd_SetDocInfo( pri, docInfo ) )	return( pdERROR );
		memcpy( &pri->pageDev, &pri->docDev, sizeof( pd_DeviceInfo ) );
	}
	/* Printer Device Restart */ 
	if( memcmp( &pri->curDev, &pri->docDev, sizeof( pd_DeviceInfo ) ) ) {
		memcpy( &pri->curDev, &pri->docDev, sizeof( pd_DeviceInfo ) );
	}
	/* State Flag Set */
	pri->docStarted = TRUE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_EndDoc
 *
 * ---------------------------------------------------------------- */
int
Ijet_EndDoc(int printerContext)
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->docStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted || pri->scnLinStarted || pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Attribute Set */
	memcpy( &pri->docDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );
	memcpy( &pri->pageDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );

	/* State Flag Set */
	pri->docStarted = FALSE;
	pri->pageStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_StartPage
 *
 * ---------------------------------------------------------------- */
int
Ijet_StartPage(int printerContext,
	       char *pageInfo)
{
	pd_PrinterInfo		*pri;
	BJPContext	*context;
	BJPControlInfo	ctrlInfo;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	if( pageInfo ) {
		pd_DebugPrint( "pageInfo = %s\n", pageInfo );
	}

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Attribute Set */
	if( pageInfo && *pageInfo ) {
		if( !pd_SetPageInfo( pri, pageInfo ) )	return( pdERROR );
	}

	/* Printer Device Restart */ 
	if( memcmp( &pri->curDev, &pri->pageDev, sizeof( pd_DeviceInfo ) ) ) {
		memcpy( &pri->curDev, &pri->pageDev, sizeof( pd_DeviceInfo ) );
		memset(&ctrlInfo, 0, sizeof(BJPControlInfo));
		setupControlInfo(&pri->pageDev, &ctrlInfo);
		context = pri->printerData;
		bjp_SetControlInfo(context, &ctrlInfo);
	}

	/* State Flag Set */
	pri->pageStarted = TRUE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_EndPage
 *
 * ---------------------------------------------------------------- */
int
Ijet_EndPage(int printerContext)
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted || pri->scnLinStarted || pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Form Feed */

	/* Attribute Set */
	memcpy( &pri->pageDev, &pri->docDev, sizeof( pd_DeviceInfo ) );

	/* State Flag Set */
	pri->pageStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_QueryColorSpace
 *
 * ---------------------------------------------------------------- */
int
Ijet_QueryColorSpace(int printerContext,
		     pdColorSpace *pcspace,
		     int *pnum )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	if( pcspace ) {
		if( *pnum < 1 ) {
			errorno = pdPARAMERROR;
			return( pdERROR );
		}
		*pcspace = pdcspaceStandardRGB;
	}
	*pnum = 1;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_SetColorSpace
 *
 * ---------------------------------------------------------------- */
int
Ijet_SetColorSpace(int printerContext,
		   pdColorSpace cspace)
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	if( cspace != pdcspaceStandardRGB ) {
		pd_DebugPrint( "pdapi:cspace = %d \n", cspace ); // For Debug!
		errorno = pdPARAMERROR;
		return( pdERROR );
	}
	gs->colorSpace = cspace;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_GetColorSpace
 *
 * ---------------------------------------------------------------- */
int
Ijet_GetColorSpace(int printerContext,
		   pdColorSpace *pcspace)
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*pcspace = gs->colorSpace;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_DrawImage
 *
 * ---------------------------------------------------------------- */
int
Ijet_DrawImage(int printerContext,
	       int sourceWidth,
	       int sourceHeight,
	       int colorDepth,
	       pdImageFormat imageFormat,
	       pdRectangle destinationSize,
	       int count,
	       void *imageData)
{
	pd_PrinterInfo		*pri;
	BJPContext	*context;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( imageFormat != pdiformatRaw ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	if( colorDepth != 24 ){
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Data Out */
	context = (BJPContext *)pri->printerData;
	if(bjp_StartDrawImage(context, sourceWidth, sourceHeight, 3 /*colorDepth*/) < 0){
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	bjp_TransferDrawImage(context, count, imageData);
	bjp_EndDrawImage(context);

	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_StartDrawImage
 *
 * ---------------------------------------------------------------- */
int
Ijet_StartDrawImage(int printerContext,
		    int sourceWidth,
		    int sourceHeight,
		    int colorDepth,
		    pdImageFormat imageFormat,
		    pdRectangle destinationSize)
{
	pd_PrinterInfo		*pri;
	BJPContext	*context;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( imageFormat != pdiformatRaw ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	if( colorDepth != 24 ){
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Data Out */
	context = (BJPContext *)pri->printerData;
	if( bjp_StartDrawImage(context, sourceWidth, sourceHeight, 3 /*colorDepth*/) < 0 ){
		errorno = pdBADREQUEST;
		return( pdERROR );
	}

	pri->drwImgStarted = TRUE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_TransferDrawImage
 *
 * ---------------------------------------------------------------- */
int
Ijet_TransferDrawImage(int printerContext,
		       int count,
		       void *imageData)
{
	pd_PrinterInfo		*pri;
	BJPContext	*context;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->drwImgStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	context = (BJPContext *)pri->printerData;
	bjp_TransferDrawImage(context, count, imageData);

	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_EndDrawImage
 *
 * ---------------------------------------------------------------- */
int
Ijet_EndDrawImage(int printerContext)
{
	pd_PrinterInfo		*pri;
	BJPContext	*context;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->drwImgStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	context = (BJPContext *)pri->printerData;
	bjp_EndDrawImage(context);

	pri->drwImgStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_StartRaster
 *
 * ---------------------------------------------------------------- */
int 
Ijet_StartRaster(int printerContext,
		 int rasterWidth)
{
	pd_PrinterInfo		*pri;
	BJPContext	*context;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* ColorSpace check */

	/* Data Out */
	context = (BJPContext *)pri->printerData;
	context->rasterdrawing = TRUE;
	if( bjp_StartDrawImage(context, rasterWidth, 1, 3) < 0 ){
		errorno = pdBADREQUEST;
		return( pdERROR );
	}

	pri->rasterStarted = TRUE;
	pri->rasterWidth = rasterWidth;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_TransferRasterData
 *
 * ---------------------------------------------------------------- */
int
Ijet_TransferRasterData(int printerContext,
			int count,
			unsigned char *data)
{
	pd_PrinterInfo		*pri;
	BJPContext	*context;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	context = (BJPContext *)pri->printerData;
	bjp_TransferDrawImage(context, count, data);

	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_SkipRaster
 *
 * ---------------------------------------------------------------- */
int
Ijet_SkipRaster(int printerContext,
		int count)
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	Ijet_EndRaster
 *
 * ---------------------------------------------------------------- */
int
Ijet_EndRaster(int printerContext)
{
	pd_PrinterInfo		*pri;
	BJPContext	*context;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	context = (BJPContext *)pri->printerData;
	bjp_EndDrawImage(context);
	context->rasterdrawing = FALSE;

	pri->rasterStarted = FALSE;
	return( pdOK );
}
