/*
 *  OpenPrinting Vector Printer Driver for Canon Pixus850i
 *  Copyright (c) 2004 E&D Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

/* $Revision: 1.9 $ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <popt.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>
#include <dlfcn.h>

#include "bjp_draw.h"

#ifdef DBGPRN
#define _dbgprint(format,...)	fprintf(stderr,format,##__VA_ARGS__)
#else
#define _dbgprint(format,...)
#endif

/*-------------------------------------------------------------*/

static long
bjp_RoundUp( double target )
{
	long	tmp_down,tmp_up;
	
	tmp_down = (long)target;
	
	if( target == (double)tmp_down )
		tmp_up = tmp_down;
	else
		tmp_up = (long)(target + 1 );
	
	return tmp_up;
	
}


/*-------------------------------------------------------------*/

/*
 * extend raster data for horizon
 */
static short
bjp_HorizonExtend(CPKByte CPKPTR	in,
		  CPKByte CPKPTR	out,
		  int		src_width,
		  int		dst_width,
		  int		component)
{
	int Quotient;
	int	Rest;
	int	TotalRest;
	int PrevPos;
	int CurrPos;
	int	x;
	unsigned char K,Y,M,C,y,m,c;
	CPKByte CPKPTR src_in;
	short result = -1;
	
	src_in = in;
	
	K = Y = M = C = y = m = c = 0;	
	Quotient  = src_width / dst_width;
	Rest      = src_width % dst_width;
	TotalRest = 0;
	PrevPos   = -1;
	CurrPos   = 0;
	
	for( x=0; x<dst_width; x++ ){
		if( CurrPos != PrevPos ){
			PrevPos = CurrPos;
			src_in = in + ( component * CurrPos );
			
			switch( component ){
			case 1:
				K = src_in[0];
				break;
			case 3:
				Y = src_in[0]; M = src_in[1];
				C = src_in[2];
				break;
			case 4:
				K = src_in[0]; Y = src_in[1];
				M = src_in[2]; C = src_in[3];
				break;
			case 6:
				K = src_in[0]; Y = src_in[1];
				M = src_in[2]; C = src_in[3];
				m = src_in[4]; c = src_in[5];
				break;
			case 7:
				K = src_in[0]; Y = src_in[1];
				M = src_in[2]; C = src_in[3];
				y = src_in[4]; m = src_in[5]; c = src_in[6];
				break;
			}
		}
		
		switch( component ){
		case 1:
			out[0] = K;
			break;
		case 3:
			out[0] = Y; out[1] = M; out[2] = C;
			break;
		case 4:
			out[0] = K; out[1] = Y; out[2] = M; out[3] = C;
			break;
		case 6:
			out[0] = K; out[1] = Y; out[2] = M;
			out[3] = C; out[4] = m; out[5] = c;
			break;
		case 7:
			out[0] = K; out[1] = Y; out[2] = M;
			out[3] = C; out[4] = y; out[5] = m; out[6] = c;
			break;
		}
		
		out += component;
		
		TotalRest += Rest;
		CurrPos += Quotient;
		
		if((( 2 * TotalRest + Rest ) >> 1) >= dst_width ){
			TotalRest -= dst_width;
			CurrPos++;
		}
	}
	
	result = 0;
	return result;
}


/*
 * whiteLine
 */
static CPKBoolean
bjp_WhiteLine(CPKByte CPKPTR buf,
	      CPKUInt32 size,
	      CPKUInt32 bpp,
	      CPKUInt32 CPKPTR left,
	      CPKUInt32 CPKPTR right)
{
	CPKUInt32 i ;
	CPKUInt32 width = size / bpp ;
	CPKByte CPKPTR src = buf ;

	*left  = 0 ;
	*right = 0 ;
	/* Check from left */
	for( i = 0 ; i < width ; i++, src += bpp ){
		if( *src != 0xff || *(src+1) != 0xff || *(src+2) != 0xff ){
			*left = i ;
			goto from_right ;
		}
	}
	*left = width ;
	return CPKTrue ;

 from_right :
	/* Check from right */
	src = buf + size - 1 ;
	for( i = 0 ; i < width ; i++, src -= bpp ){
		if( *src != 0xff || *(src-1) != 0xff || *(src-2) != 0xff ){
			*right = i ;
			break ;
		}
	}

	return CPKFalse ;
}


/*-------------------------------------------------------------*/

static void
bjp_SetupPosition(BJPContext *context,
		  long img_width,
		  long img_height,
		  BJPPos *bjppos)
{
	double	img_xres, img_yres;
	long	rightmargin,bottommargin;
	double	prn_xresolution, prn_yresolution;
	double	prnarea_width, prnarea_height;
	double	prnarea_width_org;
	double	prn_width, prn_height;
	double	mask_width, mask_height;
	double	out_width, out_height;
	double	width_blnk, height_blnk;
	double	tmp_leftskip,tmp_topskip;
	double	topmargin, leftmargin;
	BJFLTDEVICE		*dev;
	BJFLTCOLORSYSTEM	*color;
	BJFLTCOMSYSTEM		*comsys;
	CNCLPAPERSIZE		*psize;

	dev = &(context->device);
	comsys = &(context->comsystem);
	psize = &(context->papersize);
	color = &(context->color);

	memset(bjppos, 0, sizeof(BJPPos));

	bjppos->img_width = img_width;
	bjppos->img_height = img_height;

	/* set image resolution */
	bjppos->xresolution = context->resolutionX;
	bjppos->yresolution = context->resolutionY;

	/* set printer parameter */
        bjppos->topmargin       = psize->nTopMargin;
        bjppos->leftmargin      = psize->nLeftMargin;
	topmargin  = (double)((double)bjppos->topmargin  / 2540);
	leftmargin = (double)((double)bjppos->leftmargin / 2540);

	/* change resolution */
        bjppos->prn_xresolution = comsys->bjfltInputHResolution;
        bjppos->prn_yresolution = comsys->bjfltInputVResolution;

        bjppos->prn_width       = psize->nWidthPrintArea;
        bjppos->prn_height      = psize->nLengthPrintArea;

	/* set paper size */
	bjppos->PaperWidth  = psize->nSelPaperWidth;
	bjppos->PaperLength = psize->nSelPaperLength;

	rightmargin  = bjppos->PaperWidth - bjppos->prn_width - bjppos->leftmargin;
	bottommargin = bjppos->PaperLength - bjppos->prn_height - bjppos->topmargin;

	/*--- set position parameter ---*/
	bjppos->curbottommargin = (double)bottommargin;
	bjppos->currightmargin  = (double)rightmargin;
	bjppos->curtopmargin    = (double)bjppos->topmargin;
	bjppos->curleftmargin   = (double)bjppos->leftmargin;
	bjppos->curprn_width    = (double)bjppos->prn_width;
	bjppos->curprn_height   = (double)bjppos->prn_height;

	/*--- output image size ---*/
	/*--- image data ---*/
	img_xres = bjppos->xresolution;
	img_yres = bjppos->yresolution;
	
	/*--- printer data ---*/
	prn_xresolution = bjppos->prn_xresolution;
	prn_yresolution = bjppos->prn_yresolution;

	/*--- output image size [pixel] ---*/
	out_width  = (double)img_width  / img_xres * prn_xresolution;
	out_height = (double)img_height / img_yres * prn_yresolution;

	/*--- IMAGE CUT ---*/

	/*--- original printable area [inch] ---*/
	prnarea_width_org = bjppos->prn_width / 2540.0;

	/*--- current printable area ( 1/100[mm] to inch ) [inch] ---*/
	prnarea_width  = bjppos->curprn_width / 2540.0;
	prnarea_height = bjppos->curprn_height / 2540.0;

	/*--- current printable area [pixel] (with printer resolution)---*/
	prn_width  = prnarea_width * prn_xresolution;
	prn_height = prnarea_height * prn_yresolution;

	/* blank adjust */
	width_blnk  = prn_width - out_width;
	height_blnk = prn_height - out_height;
	tmp_leftskip =width_blnk / (double)2.0;
	tmp_topskip  =height_blnk / (double)2.0;

	/*--- Init mask with whole target image ---*/
	/*--- [pixel] with image resolution ---*/
	mask_width  = (double)img_width;
	mask_height = (double)img_height;

	if( out_width > prn_width ){	/* width is over */
		mask_width = (double)img_width * prn_width / out_width;
		out_width = prn_width;
		/* horizontal blank adjust */
		bjppos->width_offset += tmp_leftskip / out_width * (double)img_width * (-1.0);
	}

	if( out_height > prn_height ){	/* height is over */
		mask_height = (double)img_height * prn_height / out_height;
		out_height = prn_height;
		/* vertical blank adjust */
	}
		
	if( context->rasterdrawing){	/* raster drawing adjust */
		mask_height = (double)img_height * prn_height / out_height;
		out_height = prn_height;
		/* this adjustment valid only in the case of A4 paper size */
		if(!strcmp(context->MediaSizeName,BJP_PAPERSIZE_A4)){
			bjppos->width_offset  = leftmargin * bjppos->xresolution;
			bjppos->height_offset = topmargin * bjppos->yresolution;
		}
	}

	/*--- set parameter ---*/
	bjppos->mask_width  = mask_width;
	bjppos->mask_height = mask_height;
	bjppos->out_width   = out_width;
	bjppos->out_height  = out_height;
	bjppos->page_width  = prn_xresolution * prnarea_width_org; 
}

/*----------------------------------------------------------------------
 * public functions
 *----------------------------------------------------------------------*/

void
bjp_OutCmd(CPKByte CPKPTR buf,
	   CPKUInt32 size,
	   int prn )
{
	char *ptr;
	int w_size;
	int r_size;

	if(size == 0) return; /* do nothing. */

	ptr = buf;
	r_size = size;
	while( 1 ){
		if(r_size == 0) return;
		
		w_size = write( prn, ptr, r_size );

		if( w_size <= 0 ){
			w_size = 0;
			continue;
		}
		if( w_size < r_size ){
			r_size -= w_size;
			ptr    += w_size;
		}
		else break;
	}

	return;
}

/*----------------------------------------------------------------------
 * Raster image printing has been excuted the following functions.
 *  1) bjp_SetupImageInfo(BJPContext *, BJF_IMAGEINFO *)
 *  2) bjp_OutputImage(BJPContext *)
 *  3) bjp_FlushImage(BJPContext *)
 */

static int
bjp_SetupImageInfo(BJPContext *context,
		   long width,
		   long length,
		   short bpp)
{
	int rv = -1;
	CNCLErr  cnclerr = -1;
	CNCLData *cnclData;
	long ow;	/* out_width	*/
	long ls;	/* leftskip	*/
	long pw;	/* page_width	*/
	int (*func)();
	char *errStr;

	cnclData = context->cnclData;

	/* setup for BJPPos */
	bjp_SetupPosition(context, width, length, &context->bjppos);

	/* set image info of BJPPos to CNCLData */
	cnclData->width		= context->bjppos.page_width;
	cnclData->length	= bjp_RoundUp(context->bjppos.out_height);
	cnclData->srcdatatype	= BJP_ST_RGB24;
	cnclData->PageNum	= 1;

	context->width         = bjp_RoundUp(context->bjppos.mask_width);
	context->height        = bjp_RoundUp(context->bjppos.mask_height);
	context->width_offset  = (long)context->bjppos.width_offset;
	context->height_offset = (long)context->bjppos.height_offset;

	ow = bjp_RoundUp(context->bjppos.out_width);
	ls = context->bjppos.leftskip;
	pw = context->bjppos.page_width;
	context->ImageWidth = ((ow + ls) > pw)? pw - ls: ow;

	context->ImageHeight   = bjp_RoundUp(context->bjppos.out_height);
	context->leftskip      = context->bjppos.leftskip;
	context->topskip       = context->bjppos.topskip;
	context->page_width    = (long)context->bjppos.page_width;

#if 0
	fprintf(stderr,"\n*** set parameters ***\n");
	fprintf(stderr,"MaskWidth:%ld\n",context->width);
	fprintf(stderr,"MaskHeight:%ld\n",context->height);
	fprintf(stderr,"WidthOffset:%ld\n",context->width_offset);
	fprintf(stderr,"HeightOffset:%ld\n",context->height_offset);
	fprintf(stderr,"OutWidth:%ld\n",context->ImageWidth);
	fprintf(stderr,"OutHeight:%ld\n",context->ImageHeight);
	fprintf(stderr,"TopSkip:%ld\n",context->topskip);
	fprintf(stderr,"LeftSkip:%ld\n",context->leftskip);
	fprintf(stderr,"PageWidth[pixel]:%ld\n",context->page_width);
#endif

	context->bpp = bpp;

	/* dump RASTER */
	if(context->srcBuf == CPKNULL){
		context->srcBuf = (CPKByte CPKPTR)malloc(width * context->bpp);
	}

	if(context->rgbBuf == CPKNULL){
		context->rgbBuf = (CPKByte CPKPTR)malloc(context->page_width * context->bpp);
	}

	func = dlsym(context->dlHndl, "CNCL_SetParam");
	if((errStr = dlerror()) != NULL){
		goto onErr;
	}
	if((cnclerr = func(cnclData)) != BJPID_OK){
		fprintf(stderr,"error: CNCL_SetParam() has failed\n");
		goto onErr;
	}

	if(cnclData->Work2Buf == CPKNULL){
		cnclData->Work2Buf = (CPKVoid CPKPTR)malloc(cnclData->Work2Size);
		if(cnclData->Work2Buf == CPKNULL)
			goto onErr;
	}	

	/* START PAGE */
	func = dlsym(context->dlHndl, "CNCL_StartPage");
	if((errStr = dlerror()) != NULL){
		goto onErr;
	}
	if((cnclerr = func(cnclData)) != BJPID_OK)
		goto onErr;
	bjp_OutCmd(cnclData->outputBuffer, cnclData->outputSize, context->outfd);

	if( context->bjppos.topskip ){
		cnclData->Vpos    = context->bjppos.topskip;
		cnclData->length += context->bjppos.topskip;
	}

	cnclData->inputBuffer = context->rgbBuf ;
	cnclData->left_white  = 0 ;
	cnclData->right_white = 0 ;
	
	/* set scaling parameter */
	context->Quotient  = context->height / context->ImageHeight;
	context->Rest 	= context->height % context->ImageHeight;
	context->TotalRest = 0;
	context->PrevPos   = -1;
	context->CurrPos   = 0;

	rv = 0;

 onErr:
	return rv;
}

/*----------------------------------------------------------------------*/

static int
bjp_OutputImage(BJPContext *context)
{
	int rv = -1;
	long i;
	long tmpRestLines;
	long h;
	long restLines;
	long left=0;
	long right=0;

	long width	= context->width;
	long leftskip	= context->leftskip;
	short bpp	= context->bpp;
	long page_width	= context->page_width;
	long ImageWidth	= context->ImageWidth;
	long PrevPos	= context->PrevPos;
	long CurrPos	= context->CurrPos;
	CPKByte CPKPTR srcBuf = context->srcBuf;
	CPKByte CPKPTR rgbBuf = context->rgbBuf;
	CNCLErr  cnclerr = -1;
	CNCLData *cnclData;
	int (*func)();
	int (*fnc2)();
	char *errStr;

	if( CurrPos != PrevPos ){
		PrevPos = CurrPos;
		memset(rgbBuf, 0xff, page_width * bpp);
		if(bjp_HorizonExtend(srcBuf+(long)context->bjppos.width_offset*bpp, rgbBuf+(leftskip*bpp), width, ImageWidth, bpp) < 0)
			goto onErr;
		
	}
	
	cnclData = context->cnclData;

	/* Check white line */
	if(bjp_WhiteLine(rgbBuf, cnclData->width*bpp, bpp, &left, &right)){
		cnclData->WhiteLine = BJP_WL_WHITE;
		cnclData->left_white = left;
		cnclData->right_white = right;
	}else{
		cnclData->WhiteLine = BJP_WL_NOTWHITE;
		cnclData->left_white = left;
		cnclData->right_white = right;
	}
	
	func = dlsym(context->dlHndl, "CNCL_SourceData");
	if((errStr = dlerror()) != NULL){
		goto onErr;
	}
	restLines = func(cnclData);
	
	if(restLines < 0){
		goto onErr;
	}else if(restLines == 0){
		goto next;
	}
	
	func = dlsym(context->dlHndl, "CNCL_RasterData");
	if((errStr = dlerror()) != NULL){
		goto onErr;
	}
	fnc2 = dlsym(context->dlHndl, "CNCL_CompleteData");
	if((errStr = dlerror()) != NULL){
		goto onErr;
	}

	for( h=0; h<restLines; h++ ){
		tmpRestLines = func(cnclData);
		
		if(tmpRestLines < 0)
			goto onErr;
		
		for(i=0; i<cnclData->numBuf; i++){
			bjp_OutCmd(cnclData->rasterDataPtr[i].buf, cnclData->rasterDataPtr[i].size, context->outfd);
		}
		
		if((cnclerr = fnc2(cnclData)) != BJPID_OK)
			goto onErr;
	}
	
 next:
	cnclData->Vpos += 1 ;
	context->TotalRest += context->Rest;
	context->CurrPos += context->Quotient;
	
	if((( 2 * context->TotalRest + context->Rest ) >> 1) >= context->ImageHeight ){
		context->TotalRest -= context->ImageHeight;
		context->CurrPos++;
	}
	
	rv = 0;

 onErr:
	return(rv);
	
}


static int
bjp_FlushImage(BJPContext *context)
{
	int rv = -1;
	long restLines;
	long i;
	CNCLErr  cnclerr = -1;
	CNCLData *cnclData;
	int (*func)();
	char *errStr;

	cnclData = context->cnclData;

	/*--- Flush ---*/
	restLines = 1;
	do{
		func = dlsym(context->dlHndl, "CNCL_FlushData");
		if((errStr = dlerror()) != NULL){
			goto onErr;
		}
		restLines = func(cnclData);
		if(restLines < 0) goto onErr;

		/*--- Data ---*/
		for(i=0; i<cnclData->numBuf; i++){
			bjp_OutCmd(cnclData->rasterDataPtr[i].buf, cnclData->rasterDataPtr[i].size, context->outfd);
			}

		func = dlsym(context->dlHndl, "CNCL_CompleteData");
		if((errStr = dlerror()) != NULL){
			goto onErr;
		}
		if((cnclerr = func(cnclData)) != BJPID_OK)
			goto onErr;

	}while( restLines );

	/* ENDPAGE */
	func = dlsym(context->dlHndl, "CNCL_EndPage");
	if((errStr = dlerror()) != NULL){
		goto onErr;
	}
	if((cnclerr = func(cnclData)) != BJPID_OK)
		goto onErr;
	bjp_OutCmd(cnclData->outputBuffer, cnclData->outputSize, context->outfd);

	rv = 0;

 onErr:
	if(cnclData->Work2Buf){
		free(cnclData->Work2Buf);
		cnclData->Work2Buf = NULL;
	}
	if(context->rgbBuf){
		free(context->rgbBuf);
		context->rgbBuf = NULL;
	}
	if(context->srcBuf){
		free(context->srcBuf);
		context->srcBuf = NULL;
	}

	return(rv);
}


/*----------------------------------------------------------------------
 * public functions
 *----------------------------------------------------------------------*/

int
bjp_StartDrawImage(BJPContext *context,
		   int width,
		   int height,
		   int depth)
{
	BJImageContext *imgct = &context->img_context;
	unsigned char *p;
	int rv = -1;

	rv = bjp_SetupImageInfo(context, width, height, depth);
	if(rv < 0){
		fprintf(stderr,"bjp_SetupImageInfo() failed\n");
		goto onErr;
	}

	imgct->width = width;
	imgct->height = height;
	imgct->depth = depth;
	imgct->rassize = width * depth;
	p = (unsigned char *)malloc(width * depth);
	if(!p) goto onErr;
	imgct->rasbuf = p;
	imgct->xpos = 0;
	imgct->ypos = 0;
	rv = 0;

 onErr:
	return(rv);
}

int
bjp_TransferDrawImage(BJPContext *context,
		      int count,
		      void *data)
{
	int cnt;
	int quotient, remainder;
	int i;
	long cpos;
	BJImageContext *imgct = &context->img_context;

	_dbgprint("Trans count=%d data=%p\n", count, data);
	if(imgct->ypos < context->height_offset){
		imgct->ypos++;
		return 0;
	}

	if(imgct->xpos > 0){
		if((imgct->xpos + count) < imgct->rassize){
			memcpy(&imgct->rasbuf[imgct->xpos], data, count);
			imgct->xpos += count;
			_dbgprint("Flush(1) ypos=%d xpos=%d\n",
				imgct->ypos, imgct->xpos);
			return(0);
		}
		cnt = imgct->rassize - imgct->xpos;
		memcpy(&imgct->rasbuf[imgct->xpos], data, cnt);

		cpos = context->CurrPos;
		memcpy(context->srcBuf, imgct->rasbuf, imgct->rassize);
		while(cpos == context->CurrPos){
			_dbgprint("CurrPos<1>=%ld\n", context->CurrPos);
			bjp_OutputImage(context);
		}

		imgct->xpos = 0;
		imgct->ypos++;
		_dbgprint("Flush(2) ypos=%d xpos=%d\n",
			imgct->ypos, imgct->xpos);
		count -= cnt;
		data += cnt;
		if(count == 0)
			return(0);

		_dbgprint("Flush(2X)\n");
	}
	quotient  = count / imgct->rassize;
	remainder = count % imgct->rassize;

	for(i=0; i<quotient; i++){
		memcpy(imgct->rasbuf, data+(imgct->rassize * i), imgct->rassize);

		cpos = context->CurrPos;
		memcpy(context->srcBuf, imgct->rasbuf, imgct->rassize);
		while(cpos == context->CurrPos){
			_dbgprint("CurrPos<2>=%ld\n", context->CurrPos);
			bjp_OutputImage(context);
		}

		imgct->ypos++;
		_dbgprint("Flush(3) ypos=%d xpos=%d\n",
			imgct->ypos, imgct->xpos);
	}
	if(remainder == 0)
		return(0);
	memcpy(imgct->rasbuf, data+(imgct->rassize * i), remainder);
	imgct->xpos += remainder;
	_dbgprint("Flush(4) ypos=%d xpos=%d\n",
		imgct->ypos, imgct->xpos);

	return(0);
}

int
bjp_EndDrawImage(BJPContext *context)
{
	BJImageContext *imgct = &context->img_context;

	bjp_FlushImage(context);
	if(imgct->rasbuf)
		free(imgct->rasbuf);

	return(0);
}


